﻿//
// DirectXPage.xaml.cpp
// Implementation of the DirectXPage.xaml class.
//

#include "pch.h"
#include "DirectXPage.xaml.h"

using namespace FastZoomMap;

using namespace Concurrency;
using namespace Microsoft::WRL;
using namespace Platform;
using namespace Windows::Devices::Geolocation;
using namespace Windows::Foundation;
using namespace Windows::Foundation::Collections;
using namespace Windows::Graphics::Display;
using namespace Windows::Storage::Streams;
using namespace Windows::UI::Input;
using namespace Windows::UI::Core;
using namespace Windows::UI::Popups;
using namespace Windows::UI::Xaml;
using namespace Windows::UI::Xaml::Controls;
using namespace Windows::UI::Xaml::Controls::Primitives;
using namespace Windows::UI::Xaml::Data;
using namespace Windows::UI::Xaml::Input;
using namespace Windows::UI::Xaml::Media;
using namespace Windows::UI::Xaml::Navigation;

DirectXPage::DirectXPage() : m_renderNeeded(true)
{
    InitializeComponent();

    // Attempt to keep rotation to a minimum
    DisplayProperties::AutoRotationPreferences = DisplayProperties::NativeOrientation;

    // Install handler for Loaded event
    Loaded += ref new RoutedEventHandler(this, &DirectXPage::OnPageLoaded);

    // Create the ZoomMapRenderer
    m_renderer = ref new ZoomMapRenderer();

    // The following code was created by the template
    m_renderer->Initialize(Window::Current->CoreWindow,
                           SwapChainPanel,
                           DisplayProperties::LogicalDpi);

    Window::Current->CoreWindow->SizeChanged += 
        ref new TypedEventHandler<CoreWindow^, WindowSizeChangedEventArgs^>(this, &DirectXPage::OnWindowSizeChanged);

    DisplayProperties::LogicalDpiChanged +=
        ref new DisplayPropertiesEventHandler(this, &DirectXPage::OnLogicalDpiChanged);

    DisplayProperties::OrientationChanged +=
        ref new DisplayPropertiesEventHandler(this, &DirectXPage::OnOrientationChanged);

    DisplayProperties::DisplayContentsInvalidated +=
        ref new DisplayPropertiesEventHandler(this, &DirectXPage::OnDisplayContentsInvalidated);

    m_eventToken = CompositionTarget::Rendering::add(ref new EventHandler<Object^>(this, &DirectXPage::OnRendering));

    m_timer = ref new BasicTimer();
}

void DirectXPage::OnPageLoaded(Object^ sender, RoutedEventArgs^ args)
{
    // Create a Geolocator object
    geolocator = ref new Geolocator();

    // Get the current location
    task<Geoposition^> getGeopositionTask = create_task(geolocator->GetGeopositionAsync());
    getGeopositionTask.then([this] (task<Geoposition^> thisTask)
    {
        Geoposition^ geopos = nullptr;

        try
        {
            geopos = thisTask.get();
        }
        catch (Exception^)
        {
        }

        if (geopos == nullptr)
        {
            MessageDialog^ msgdlg = ref new MessageDialog("Program needs access to computer's location");
            task<IUICommand^> showMsgTask = create_task(msgdlg->ShowAsync());
            showMsgTask.then([] (IUICommand^ command)
            {
            });
        }
        else
        {
            // Give renderer the geographical position
            m_renderer->SetGeolocation(geopos->Coordinate->Longitude, 
                                       geopos->Coordinate->Latitude);

            // Access Bing Maps to get metadata
            task<ViewParams> getViewParamsTask = create_task(GetViewParamsAsync(L"Aerial"));
            getViewParamsTask.then([this] (ViewParams viewParams)
            {
                // Set the Slider minimum and maximum
                zoomLevelSlider->Minimum = viewParams.minZoomLevel;
                zoomLevelSlider->Maximum = viewParams.maxZoomLevel;

                // Set the "powered by Bing" bitmap
                poweredByBitmap->UriSource = ref new Uri(ref new String(viewParams.PoweredByBingUri.data()));
                poweredByDisplay->Visibility = Windows::UI::Xaml::Visibility::Visible;

                // Pass the view parameters to the renderer
                m_renderer->SetViewParams(ref new String(viewParams.UriTemplate.data()), 
                                          viewParams.minZoomLevel, 
                                          viewParams.maxZoomLevel);
            });
        }
    });
}

task<ViewParams> DirectXPage::GetViewParamsAsync(wchar_t* type)
{
    // Define the URI to get Bing imagery metadata
    wchar_t* bingCredentials = L"insert your own Bing Maps key here!";
 
    std::wstring strUri(L"http://dev.virtualearth.net/REST/V1/Imagery/Metadata/");
    strUri.append(type);
    strUri.append(L"?output=xml&key=");
    strUri.append(bingCredentials);
    Uri^ uri = ref new Uri(ref new String(strUri.c_str()));

    // Open the URI for reading
    RandomAccessStreamReference^ streamRef = RandomAccessStreamReference::CreateFromUri(uri);
    task<IRandomAccessStreamWithContentType^> openReadTask = create_task(streamRef->OpenReadAsync());
    return openReadTask.then([this](IRandomAccessStreamWithContentType^ stream)
    {
        // Create a COM IStream object from the IRandomAccessStreamWithContentType
        ComPtr<IStream> pStream;
        HRESULT hr = CreateStreamOverRandomAccessStream(stream, __uuidof(IStream), &pStream);

        if (FAILED(hr))
            throw ref new COMException(hr, "CreateStreamOverRandomAccessStream failure");
        
        // Create a COM IXmlReader object
        ComPtr<IXmlReader> pXmlReader;
        hr = CreateXmlReader(__uuidof(IXmlReader), &pXmlReader, nullptr);

        if (FAILED(hr))
            throw ref new COMException(hr, "CreateXmlReader failure");

        hr = pXmlReader->SetInput(pStream.Get());

        if (FAILED(hr))
            throw ref new COMException(hr, "XmlReader->SetInput failure");

        // Look for five elements
        bool awaitingZoomMin = false;
        bool awaitingZoomMax = false;
        bool awaitingSubDomains = false;
        bool awaitingImageUrl = false;
        bool awaitingPoweredByBingUri = false;

        // Store the results here
        ViewParams viewParams;
        std::vector<std::wstring> subdomains;

        // Look until the Read doesn't return any more nodes
        XmlNodeType xmlNodeType;
        while (S_OK == (hr = pXmlReader->Read(&xmlNodeType)))
        {
            // Check for beginning and end of elements
            if (xmlNodeType == XmlNodeType_Element || xmlNodeType == XmlNodeType_EndElement)
            {
                LPCWSTR pName;
                unsigned int charCount;
                hr = pXmlReader->GetLocalName(&pName, &charCount);

                if (FAILED(hr))
                    throw ref new COMException(hr, "XmlReader->GetLocalName failure");

                if (wcscmp(pName, L"ImageUrl") == 0)
                {
                    awaitingImageUrl = (xmlNodeType == XmlNodeType_Element);
                }
                else if (wcscmp(pName, L"ImageUrlSubdomains") == 0)
                {
                    awaitingSubDomains = (xmlNodeType == XmlNodeType_Element);
                }
                else if (wcscmp(pName, L"ZoomMin") == 0)
                {
                    awaitingZoomMin = (xmlNodeType == XmlNodeType_Element);
                }
                else if (wcscmp(pName, L"ZoomMax") == 0)
                {
                    awaitingZoomMax = (xmlNodeType == XmlNodeType_Element);
                }
                else if (wcscmp(pName, L"BrandLogoUri") == 0)
                {
                    awaitingPoweredByBingUri = (xmlNodeType == XmlNodeType_Element);
                }
            }

            // Get value of element
            if (xmlNodeType == XmlNodeType_Text)
            {
                LPCWSTR pValue = nullptr;
                unsigned int charCount = 0;
                hr = pXmlReader->GetValue(&pValue, &charCount);

                if (FAILED(hr))
                    throw ref new COMException(hr, "XmlReader->GetValue failure");

                if (awaitingImageUrl)
                {
                    viewParams.UriTemplate = std::wstring(pValue);
                }
                else if (awaitingSubDomains)
                {
                    subdomains.push_back(std::wstring(pValue));
                }
                else if (awaitingZoomMin)
                {
                    viewParams.minZoomLevel = _wtoi(pValue);
                }
                else if (awaitingZoomMax)
                {
                    viewParams.maxZoomLevel = _wtoi(pValue);
                }
                else if (awaitingPoweredByBingUri)
                {
                    viewParams.PoweredByBingUri = std::wstring(pValue);
                }
            }
        }

        // Need to make some modifications to the URI string
        int pos = viewParams.UriTemplate.find(L"{culture}");

        if (pos != std::wstring::npos)
            viewParams.UriTemplate.replace(pos, 9, L"en-us");

        pos = viewParams.UriTemplate.find(L"{subdomain}");

        if (pos != std::wstring::npos)
            viewParams.UriTemplate.replace(pos, 11, subdomains[0].c_str());

        return viewParams;
    });
}

// The next four methods are part of the standard template
void DirectXPage::OnWindowSizeChanged(CoreWindow^ sender, WindowSizeChangedEventArgs^ args)
{
	m_renderer->UpdateForWindowSizeChange();
	m_renderNeeded = true;
}

void DirectXPage::OnLogicalDpiChanged(Object^ sender)
{
	m_renderer->SetDpi(DisplayProperties::LogicalDpi);
	m_renderNeeded = true;
}

void DirectXPage::OnOrientationChanged(Object^ sender)
{
	m_renderer->UpdateForWindowSizeChange();
	m_renderNeeded = true;
}

void DirectXPage::OnDisplayContentsInvalidated(Object^ sender)
{
	m_renderer->ValidateDevice();
	m_renderNeeded = true;
}

// Only change from template: call to NeedsRender method in renderer
void DirectXPage::OnRendering(Object^ sender, Object^ args)
{
    m_renderNeeded |= m_renderer->NeedsRender();

    if (m_renderNeeded)
    {
        m_timer->Update();
        m_renderer->Update(m_timer->Total, m_timer->Delta);
        m_renderer->Render();
        m_renderer->Present();
        m_renderNeeded = false;
    }
}

// Part of the standard template, but calls to renderer methods removed
void DirectXPage::SaveInternalState(IPropertySet^ state)
{
}

void DirectXPage::LoadInternalState(IPropertySet^ state)
{
}

// Handle events from the Slider
void DirectXPage::OnSliderValueChanged(Object^ sender, 
                                       RangeBaseValueChangedEventArgs^ args)
{
    m_renderer->SetZoomLevel(args->NewValue);
}
