﻿#include "pch.h"
#include "ZoomMapRenderer.h"

using namespace Concurrency;
using namespace D2D1;
using namespace DirectX;
using namespace Microsoft::WRL;
using namespace Platform;
using namespace Windows::Foundation;
using namespace Windows::Foundation::Collections;
using namespace Windows::Storage;
using namespace Windows::Storage::Streams;
using namespace Windows::UI::Core;

ZoomMapRenderer::ZoomMapRenderer() : integerLongitude(-1),
                                     integerLatitude(-1),
                                     fractionalZoomLevel(1.0),
                                     zoomLevelCount(0),
                                     horzTileCount(0),
                                     vertTileCount(0),
                                     pBitmapTileArray(nullptr),
                                     pPixelOffsetArray(nullptr),
                                     uriTemplate(L""),
                                     m_renderNeeded(true)
{
}

ZoomMapRenderer::~ZoomMapRenderer()
{
    if (pPixelOffsetArray != nullptr)
        delete[] pPixelOffsetArray;

    if (pBitmapTileArray != nullptr)
    {
        for (int i = 0; i < zoomLevelCount * horzTileCount * vertTileCount; i++)
            if (pBitmapTileArray[i] != nullptr)
                pBitmapTileArray[i]->Release();

        delete[] pBitmapTileArray;
    }
}

void ZoomMapRenderer::CreateDeviceIndependentResources()
{
	DirectXBase::CreateDeviceIndependentResources();
}

void ZoomMapRenderer::CreateDeviceResources()
{
	DirectXBase::CreateDeviceResources();
}

void ZoomMapRenderer::CreateWindowSizeDependentResources()
{
    DirectXBase::CreateWindowSizeDependentResources();

    // A possible enhancement is to re-do the bitmap array when 
    //      the display size changes.

    if (horzTileCount > 0 && vertTileCount > 0)
        return;

    horzTileCount = 1 + 2 * int(ceil(m_windowBounds.Width / 2 / 128));
    vertTileCount = 1 + 2 * int(ceil(m_windowBounds.Height / 2 / 128));

    // Remove the following code to fill the window
    horzTileCount = 9;      // width of 1024
    vertTileCount = 7;      // height of 768

    DownloadAllMapTiles();
}

bool ZoomMapRenderer::NeedsRender()
{
    bool renderNeeded = m_renderNeeded;
    m_renderNeeded = false;
    return renderNeeded;
}

void ZoomMapRenderer::SetGeolocation(double longitude, double latitude)
{
    // A possible enhancement is to to re-do the bitmap array when
    //      the longitude and latitude changes.

    if (integerLongitude != -1 && integerLatitude != -1)
        return;

    // Calculate integer longitude and latitude
    double pi = 3.14159;
    double relativeLongitude = (180 + longitude) / 360;
    integerLongitude = (int)(relativeLongitude * (1 << BITRES));

    double sineTerm = sin(pi * latitude / 180);
    double relativeLatitude = 0.5 - log((1 + sineTerm) / (1 - sineTerm)) / (4 * pi);
    integerLatitude = (int)(relativeLatitude * (1 << BITRES));

    DownloadAllMapTiles();
}

void ZoomMapRenderer::SetZoomLevel(double zoomLevel)
{
    fractionalZoomLevel = zoomLevel;
    m_renderNeeded = true;
}

void ZoomMapRenderer::SetViewParams(String^ uriTemplate, int minZoomLevel, int maxZoomLevel)
{
    this->uriTemplate = std::wstring(uriTemplate->Data());
    this->minZoomLevel = minZoomLevel;
    this->maxZoomLevel = maxZoomLevel;
    zoomLevelCount = maxZoomLevel - minZoomLevel + 1;

    DownloadAllMapTiles();
}

void ZoomMapRenderer::DownloadAllMapTiles()
{
    // Check if we're already downloaded
    if (pBitmapTileArray != nullptr)
        return;

    // Check if we have the tile URI template
    if (uriTemplate.length() == 0)
        return;

    // This should also be fine by this point
    if (zoomLevelCount == 0)
        return;

    // Check if we have a geographical position
    if (integerLatitude == -1 || integerLongitude == -1)
        return;

    // Check if we have the window size 
    if (horzTileCount == 0 || vertTileCount == 0)
        return;

    // Ready to go! Allocate two arrays
    pBitmapTileArray = new ID2D1Bitmap*[zoomLevelCount * vertTileCount * horzTileCount];
    pPixelOffsetArray = new D2D1_POINT_2U[zoomLevelCount];

    // Loop through the zoom levels
    for (int zoomLevel = minZoomLevel; zoomLevel <= maxZoomLevel; zoomLevel++)
    {
        int zoomIndex = zoomLevel - minZoomLevel;   // zero based

        // Get coordinates and pixel offsets based on zoom level
        int croppedLongitude = integerLongitude >> (BITRES - zoomLevel);
        int croppedLatitude = integerLatitude >> (BITRES - zoomLevel);
        int xPixelOffset = (integerLongitude >> (BITRES - zoomLevel - 8)) % 256;
        int yPixelOffset = (integerLatitude >> (BITRES - zoomLevel - 8)) % 256;

        // Save that pixel offset as a Point
        pPixelOffsetArray[zoomIndex] = Point2U(xPixelOffset, yPixelOffset);

        // This is the maximum limit of an integer geographical location
        int maxValue = (1 << zoomLevel) - 1;

        // Loop through the rows of tiles
        for (int row = -vertTileCount / 2; row <= vertTileCount / 2; row++)
        {
            int rowIndex = row + vertTileCount / 2;

            // Loop through the columns of tiles
            for (int col = -horzTileCount / 2; col <= horzTileCount / 2; col++)
            {
                int colIndex = col + horzTileCount / 2;
                int arrayIndex = (zoomIndex * vertTileCount + rowIndex) * horzTileCount + colIndex;

                // Initialize the bitmap tile array cell
                pBitmapTileArray[arrayIndex] = nullptr;

                // Check if we've gone beyond the bounds
                if (croppedLongitude + col < 0 ||
                    croppedLongitude + col > maxValue ||
                    croppedLatitude + row < 0 ||
                    croppedLatitude + row > maxValue)
                {
                    continue;
                }

                // Calculate a quadkey
                int longitude = croppedLongitude + col;
                int latitude = croppedLatitude + row;
                std::wstring quadkey = ToQuadKey(longitude, latitude, zoomLevel);

                // Determine tile URI
                std::wstring uri(uriTemplate);
                int pos = uri.find(L"{quadkey}");
                uri.replace(pos, 9, quadkey);

                // Extract filename for caching
                pos = uri.rfind('/') + 1;
                int endPos = uri.find('?', pos);
                std::wstring filename;

                if (endPos != std::wstring::npos)
                    filename = uri.substr(pos, endPos - pos);
                else
                    filename = uri.substr(pos, std::wstring::npos);

                // Check if the required tile is stored in local storage
                StorageFolder^ localFolder = ApplicationData::Current->LocalFolder;
                task<StorageFile^> getFileTask = 
                    create_task(localFolder->GetFileAsync(ref new String(filename.data())));
                getFileTask.then([this, arrayIndex, filename, uri] (task<StorageFile^> thisTask)
                {
                    StorageFile^ storageFile = nullptr;

                    try
                    {
                        storageFile = thisTask.get();
                    }
                    catch (Exception^)
                    {
                    }

                    task<ID2D1Bitmap *> loadBitmapTask;

                    // If tile is in local storage, load it
                    if (storageFile != nullptr)
                    {
                        loadBitmapTask = create_task(LoadBitmap(storageFile, nullptr));
                    }
                    // Otherwise, download it and save it to local storage
                    else
                    {
                        // Open the URI for reading
                        Uri^ uriTile = ref new Uri(ref new String(uri.data()));
                        RandomAccessStreamReference^ streamRef = 
                            RandomAccessStreamReference::CreateFromUri(uriTile);
                        loadBitmapTask = create_task(LoadBitmap(streamRef, 
                                                                ref new String(filename.data())));
                    }

                    loadBitmapTask.then([this, arrayIndex] (ID2D1Bitmap * bitmap)
                    {
                        // Save the bitmap in the array
                        pBitmapTileArray[arrayIndex] = bitmap;

                        // Refresh the screen
                        m_renderNeeded = true;
                    });
                });
            }
        }
    }
}

task<ID2D1Bitmap*> ZoomMapRenderer::LoadBitmap(IRandomAccessStreamReference^ streamRef, 
                                               String^ saveFilename)
{
    task<IRandomAccessStreamWithContentType^> openReadTask = 
                                                create_task(streamRef->OpenReadAsync());
    return openReadTask.then([this, saveFilename](IRandomAccessStreamWithContentType^ stream)
    {
        // Check for a zero size, and avoid dealing with it
        if (stream->Size == 0)
            return (ID2D1Bitmap*) nullptr;

        // Possibly save the file to local storage
        if (saveFilename != nullptr)
        {
            SaveBitmapFile(stream->CloneStream(), saveFilename);
        }

        // Create a COM IStream object from the IRandomAccessStreamWithContentType
        ComPtr<IStream> pStream;
        HRESULT hr = CreateStreamOverRandomAccessStream(stream, IID_PPV_ARGS(&pStream));

        if (FAILED(hr))
            throw ref new COMException(hr, "CreateStreamOverRandomAccessStream failure");

        // Create a WIC Stream object from the IStream object
        ComPtr<IWICStream> pWicStream;
        hr = m_wicFactory->CreateStream(&pWicStream);

        if (FAILED(hr))
            throw ref new COMException(hr, "CreateStream failure");

        pWicStream->InitializeFromIStream(pStream.Get());
        
        // Create a decoder based on the bitmap stream
        ComPtr<IWICBitmapDecoder> pDecoder;
        hr = m_wicFactory->CreateDecoderFromStream(pWicStream.Get(), nullptr, 
                                                   WICDecodeMetadataCacheOnDemand, 
                                                   &pDecoder);
        if (FAILED(hr))
            throw ref new COMException(hr, "CreateDecoderFromStream failure");

        // Get the first (and only) frame
        ComPtr<IWICBitmapFrameDecode> pFrame;
        hr = pDecoder->GetFrame(0, &pFrame);

        if (FAILED(hr))
            throw ref new COMException(hr, "GetFrame failure");

        // Create a converter to obtain the correct color format
        ComPtr<IWICFormatConverter> pConverter;
        hr = m_wicFactory->CreateFormatConverter(&pConverter);

        if (FAILED(hr))
            throw ref new COMException(hr, "CreateFormatConverter failure");

        hr = pConverter->Initialize(pFrame.Get(), GUID_WICPixelFormat32bppPBGRA, 
                                    WICBitmapDitherTypeNone, nullptr, 0.0f, 
                                    WICBitmapPaletteTypeCustom);

        if (FAILED(hr))
            throw ref new COMException(hr, "pConverter->Initialize failure");

        // Now create a Direct2D bitmap from the WIC bitmap
        ID2D1Bitmap * bitmap;
        hr = m_d2dContext->CreateBitmapFromWicBitmap(pConverter.Get(), &bitmap);

        if (FAILED(hr))
            throw ref new COMException(hr, "CreateBitmapFromWicBitmap failure");

        return bitmap;
    });
}

void ZoomMapRenderer::SaveBitmapFile(IRandomAccessStream^ inpStream, String^ filename)
{
    StorageFolder^ localFolder = ApplicationData::Current->LocalFolder;
    task<StorageFile^> createFileTask = 
        create_task(localFolder->CreateFileAsync(filename, 
                                                 CreationCollisionOption::ReplaceExisting));
    createFileTask.then([inpStream] (StorageFile^ storageFile)
    {
        task<IRandomAccessStream^> openTask = 
            create_task(storageFile->OpenAsync(FileAccessMode::ReadWrite));
        openTask.then([inpStream] (IRandomAccessStream^ outStream)
        {
            int size = (int)inpStream->Size;
            IBuffer^ buffer = ref new Buffer(size);

            task<IBuffer^> readTask = 
                create_task(inpStream->ReadAsync(buffer, size, InputStreamOptions::None));
            readTask.then([outStream] (IBuffer^ buffer)
            {
                task<size_t> writeTask = create_task(outStream->WriteAsync(buffer));
                writeTask.then([outStream] (size_t count)
                {
                    task<bool> flushTask = create_task(outStream->FlushAsync());
                    flushTask.then([] (bool success)
                    {
                    });
                });
            });
        });
    });
}

std::wstring ZoomMapRenderer::ToQuadKey(int longitude, int latitude, int level)
{
    int64 quadkey = 0;
    int mask = 1 << (level - 1);

    for (int i = 0; i < level; i++)
    {
        quadkey <<= 2;

        if ((longitude & mask) != 0)
            quadkey |= 1;

        if ((latitude & mask) != 0)
            quadkey |= 2;

        mask >>= 1;
    }

    std::wstring str;

    for (int i = 0; i < level; i++)
    {
        str.insert(0, 1, (quadkey & 3) + '0');
        quadkey >>= 2;
    }
    return str;
}


void ZoomMapRenderer::Update(float timeTotal, float timeDelta)
{
	(void) timeTotal; // Unused parameter.
	(void) timeDelta; // Unused parameter.

	// Add code to update time dependent objects here.
}

void ZoomMapRenderer::Render()
{
    m_d2dContext->BeginDraw();
    m_d2dContext->Clear(ColorF(ColorF::CornflowerBlue));

    if (pBitmapTileArray != nullptr)
    {
        // Remove these three statements (and Pop) to display on the whole window
        float left = (m_windowBounds.Width - 1024) / 2;
        float top = (m_windowBounds.Height - 1024) / 2;
        m_d2dContext->PushAxisAlignedClip(RectF(left, top, left + 1024, top + 1024), 
                                          D2D1_ANTIALIAS_MODE_PER_PRIMITIVE);

        double fractionalZoomIndex = fractionalZoomLevel - minZoomLevel;
        float fractionalPart = float(fmod(fractionalZoomIndex, 1));

        // Calculate opacity for map tile on top
        float opacity = max(0, min(1, 2 * (fractionalPart - 0.25f)));

        int zoomIndex1 = int(fractionalZoomIndex);
        float scale1 = float(pow(2, fractionalPart));

        int zoomIndex2 = zoomIndex1 + 1;
        float scale2 = float(scale1 / 2);

        if (opacity < 1)
            DrawTileArray(zoomIndex1, scale1, 1);

        if (opacity > 0 && fractionalZoomLevel < maxZoomLevel)
            DrawTileArray(zoomIndex2, scale2, opacity);

        m_d2dContext->PopAxisAlignedClip();
    }

    // Ignore D2DERR_RECREATE_TARGET. This error indicates that the device
    // is lost. It will be handled during the next call to Present.
    HRESULT hr = m_d2dContext->EndDraw();
    if (hr != D2DERR_RECREATE_TARGET)
    {
        DX::ThrowIfFailed(hr);
    }

    m_renderNeeded = false;
}

void ZoomMapRenderer::DrawTileArray(int zoomIndex, float scale, float opacity)
{
    D2D_POINT_2U pixelOffset = pPixelOffsetArray[zoomIndex];

    for (int rowIndex = 0; rowIndex < vertTileCount; rowIndex++)
    {
        int row = rowIndex - vertTileCount / 2;

        for (int colIndex = 0; colIndex < horzTileCount; colIndex++)
        {
            int col = colIndex - horzTileCount / 2;

            int index = (zoomIndex * vertTileCount + rowIndex) * horzTileCount + colIndex;
            ID2D1Bitmap* bitmap = pBitmapTileArray[index];

            if (bitmap != nullptr)
            {
                float x = int(m_windowBounds.Width) / 2 + scale * (256 * col - float(pixelOffset.x));
                float y = int(m_windowBounds.Height) / 2 + scale * (256 * row - float(pixelOffset.y));

                D2D1_RECT_F rect = RectF(x, y, x + scale * 256, y + scale * 256);
                m_d2dContext->DrawBitmap(bitmap, &rect, opacity);
            }
        }
    }
}


